
/*
 * Removal of this header is illegal.
 * Written by Scott Auge scott_auge@yahoo.com sauge@amduus.com
 * Copyright (c) 2003 Amduus Information Works, Inc.  www.amduus.com
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Amduus Information Works
 *      Inc. and its contributors.
 * 4. Neither the name of Amduus Information Works, Inc. nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY AMDUUS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AMDUUS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

DEF VAR RCSVersion AS CHARACTER INIT "$Id$" NO-UNDO. 

DEFINE INPUT PARAMETER cWord     AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cDictName AS CHARACTER NO-UNDO.
DEFINE OUTPUT PARAMETER cList    AS CHARACTER NO-UNDO.

DEFINE VARIABLE dDistance     AS DECIMAL NO-UNDO.
DEFINE VARIABLE dThreshhold   AS DECIMAL NO-UNDO.

/*************************************************************************/
/* Got to get at least 50% of the word to match before recommending it.  */
/* This is the threshhold.  It is learned hard by testing.  It is a      */
/* "magic" number.   If you wish to adjust it, prepare to offer up a     */
/* dead chicken dance to the gods.                                       */
/*************************************************************************/

ASSIGN dThreshhold = 0.11128.

/*************************************************************************/
/* These distances are the real secret of the stew.  If the right one    */
/* can be figured out, then the spell checker will work wonders!         */
/* Another option might be to switch word to Metaphone equivalent and    */
/* search on that.                                                       */
/*************************************************************************/

/*************************************************************************/
/* Can use hamming distance only on two words with same count of letters */
/* Returns a ? when it cannot be used.                                   */
/*************************************************************************/

FUNCTION HammingDistance RETURNS INTEGER
(INPUT cWord1 AS CHARACTER, INPUT cWord2 AS CHARACTER):

  DEFINE VARIABLE i AS INTEGER NO-UNDO.
  DEFINE VARIABLE l AS INTEGER NO-UNDO.
  DEFINE VARIABLE c AS INTEGER NO-UNDO.

  ASSIGN l = LENGTH(cWord1).
  IF LENGTH(cWord2) <> l THEN RETURN ?.

  ASSIGN c = 0.
  DO i = 1 TO l:
    IF SUBSTRING(cWord1, i, 1) <> SUBSTRING(cWord, i, 1) THEN
     ASSIGN c = c + 1.
  END. /* DO */
 
  RETURN c.
 
END. /* FUNCTION HammingDistance () */

/*************************************************************************/
/* Auge Distance.  Haven't seen a distance like this defined, so we      */
/* will try this one out.  Distance = Difference in count between        */
/* letters + Difference in placement of letters.                         */
/*************************************************************************/

FUNCTION AugeDistance RETURNS INTEGER
(INPUT cWord1 AS CHARACTER, INPUT cWord2 AS CHARACTER):

  DEFINE VARIABLE iCount      AS INTEGER NO-UNDO.
  DEFINE VARIABLE cShrtWord   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE cLongWord   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE cTempWord   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE iOffset     AS INTEGER NO-UNDO.
  DEFINE VARIABLE iOffset1    AS INTEGER NO-UNDO.
  DEFINE VARIABLE iMaxOffset  AS INTEGER NO-UNDO.
  DEFINE VARIABLE iMaxOffset1 AS INTEGER NO-UNDO.
  DEFINE VARIABLE cOffsetChar AS CHARACTER NO-UNDO.

  /****************************************************************/
  /* If the lengths between the long word and the short word are  */
  /* just to far apart, then don't even bother, return 1.         */
  /****************************************************************/

  ASSIGN cShrtWord = IF (LENGTH(cWord1) < LENGTH(cWord2)) THEN cWord1 ELSE cWord2.
  ASSIGN cLongWord = IF (LENGTH(cWord1) > LENGTH(cWord2)) THEN cWord1 ELSE cWord2.

  /****************************************************************/
  /* Note this number is also a "magic" number.                   */
  /****************************************************************/

  IF LENGTH(cShrtWord) / LENGTH(cLongWord) < .865 THEN RETURN ?.

  /****************************************************************/
  /* Try to sprinkle the short words letters across the long word */
  /* If most of the long word is replaced by the letters of the   */
  /* short word, then the words are candidates for each other.    */
  /****************************************************************/

  ASSIGN iMaxOffset = LENGTH(cShrtWord).
  ASSIGN iMaxOffSet1 = LENGTH(cLongWord).

  LOOPOFFSET:
  DO iOffset = 1 TO iMaxOffset:
    ASSIGN cOffsetChar = SUBSTRING(cShrtWord, iOffset, 1).

    DO iOffset1 = 1 TO iMaxOffSet1:

    ASSIGN cLongWord = REPLACE(cLongWord, cOffsetChar, "").

    END. /* DO iOffset1 */
  END. /* DO iOffset */

  ASSIGN iCount = LENGTH(cLongWord).

  /****************************************************************/
  /* That's all folks!                                            */
  /****************************************************************/

  RETURN iCount.

END. /* FUNCTION AugeDistance() */

/*************************************************************************/
/* Computes the distance between two words                               */
/*************************************************************************/

FUNCTION Distance RETURNS DECIMAL 
(INPUT cWord1 AS CHARACTER, cWord2 AS CHARACTER):

  DEFINE VARIABLE dDistance        AS DECIMAL NO-UNDO.
  DEFINE VARIABLE dAugeDistance    AS DECIMAL NO-UNDO.
  DEFINE VARIABLE dHammingDistance AS DECIMAL NO-UNDO.
  DEFINE VARIABLE cLongWord        AS CHARACTER NO-UNDO.

  ASSIGN cLongWord = IF LENGTH(cWord1) > LENGTH(cWord2) THEN cWord1 ELSE cWord2.

  ASSIGN dAugeDistance = AugeDistance(cWord1, cWord2) / LENGTH(cLongWord).
  ASSIGN dHammingDistance = HammingDistance(cWord1, cWord2) / LENGTH(cLongWord).

  IF dHammingDistance <> ? THEN DO:
    ASSIGN dDistance = IF dAugeDistance > dHammingDistance THEN dAugeDistance 
                       ELSE dHammingDistance.
  END.
  ELSE 
    ASSIGN dDistance = dAugeDistance.

  RETURN dDistance.

END. /* FUNCTION Distance */

/*******************************************************************/
/* Found a spelling easily, just return the word                   */
/*******************************************************************/

IF CAN-FIND(Spelling WHERE Spelling.Word = cWord AND Spelling.DictName = cDictName) THEN DO:
  ASSIGN cList = cWord.
  RETURN.
END. /* IF CAN-FIND */

/**********************************************************************/
/* Ah, we need to find a "distance" from the word to likely spellings */
/* Try to assume the user entered the first letter of the word cor-   */
/* rectly.                                                            */
/**********************************************************************/

FOR EACH Spelling NO-LOCK
WHERE Spelling.DictName = cDictName
  AND Spelling.Word BEGINS SUBSTRING(cWord, 1, 1):

  ASSIGN dDistance = Distance(Spelling.Word, cWord).
  IF dDistance < dThreshhold THEN ASSIGN cList = cList + "," + Spelling.Word.

  IF LENGTH(cList) > 3000 THEN LEAVE.

  /* MESSAGE Spelling.Word dDistance dThreshhold. */

END. /* FOR EACH Spelling */

ASSIGN cList = SUBSTRING(cList, 2, LENGTH(cList) - 1).

